/////////////////////////////////////////////////////////
// memory hash table to be used in hash join.
///////////////////////////////////////////////////////

#include "sm_hash.h"

hash_table_t::hash_table_t (uint4_t size)
{
    m_size = size;
    m_table = new table_entry_t [m_size];
}

hash_table_t::~hash_table_t()
{
    this->reset();
    delete [] m_table;
}

uint4_t hash_table_t::hash_val(const char* key)
{ return hash(key) % m_size; }

void hash_table_t::add(const char * key, void* val)
{
    uint4_t hval = hash_val(key);
    table_entry_t*  bucket = &m_table[hval];

    // see if the key is already in the table.
    table_entry_t* entry = NULL;
    for (entry=bucket->next; entry!=bucket; entry = entry->next)
        if (strcmp(key, entry->key) == 0) break;
  
    if (entry == bucket) {
        // the key is not in the bucket.
        entry = new table_entry_t;
        int len = strlen(key);
        entry->key = new char[len+1];
        strncpy(entry->key, key,len);
        entry->key[len] = '\0';

        entry->next = bucket->next;
        entry->prev = bucket;
        bucket->next->prev = entry;
        bucket->next= entry;
    }

    entry->val_array.push_back(val);
}


hash_table_t::val_array_t* hash_table_t::find_match(const char* key)
{
    uint4_t hval = hash_val(key);
    table_entry_t* bucket = &m_table[hval];
    table_entry_t* entry;
  
    for (entry = bucket->next; entry!=bucket; entry = entry->next)
        if (strcmp(key, entry->key) == 0) break;
  
    // find the match
    if (entry != bucket)
        return &entry->val_array;
    else // not found
        return NULL;
}
  
void hash_table_t::reset()
{
    table_entry_t* entry;
    uint4_t bukInx;

    for (bukInx =0; bukInx < m_size; bukInx++) 
    {
    	for (entry = m_table[bukInx].next; 
	     entry != &m_table[bukInx]; 
	     entry=entry->next) 
	{
	    w_assert3(entry!=NULL);
	    /* I don't think this does anything and I am not sure why it's here anyway
    	    for (i=0; i < entry->val_array.getsize(); i++)
	        delete entry->val_array[i];
    	    entry->val_array.reset();
	    */
    	    delete [] entry->key;
    	    entry->prev->next = entry->next;
    	    entry->next->prev = entry->prev;
    	    delete entry;
        }
    }
}

uint4_t hash_func_c::operator()(const file_info_t& finfo, const char* frec, int col)
{

    const char* field_ptr = FIELD_START(finfo,frec,col);

    switch(finfo.field_type[col]) 
    {
      case SM_INTEGER:
        return this->operator()(*(int*)field_ptr);
      
      case SM_FLOAT:
        return this->operator()(*(double*)field_ptr);

      case SM_DATE:
      case SM_CHAR:
      case SM_VARCHAR:
        // string is terminated
        return this->operator()(field_ptr);
      default:
        cerr << "Bad type found in column " << col << endl;
        w_assert3(0);
        return 0;
    }
}

uint4_t hash_func_c::operator()(const file_info_t& finfo, const precord_t& prec, 
				int col)
{
  switch(finfo.field_type[col]) 
    {
    case SM_INTEGER:
      return this->operator()(*(int*) prec.data[col]);
      
    case SM_FLOAT:
      return this->operator()(*(double*) prec.data[col]);

    case SM_DATE:
    case SM_CHAR:
    case SM_VARCHAR:
      return this->operator()(prec.data[col]) ;
      
    default:
      cerr << "Bad type found in column " << col << endl;
      w_assert3(0) ;
      return 0 ;
    }
}

uint4_t hash_func_c::operator()(int int_val)
{
  sprintf(buf, "%d", int_val) ;
  return this->operator()(buf) ;
}

uint4_t hash_func_c::operator()(double double_val)
{
  sprintf(buf, "%f", double_val);
  return this->operator()(buf) ;
}

uint4_t hash_func_c::operator()(const char* str_val)
{
  uint4_t h = 0;
  for ( ; *str_val; ++str_val) h = 5*h + *str_val;
  return h % ht_size ;
}

uint4_t hash_func_c::operator()(char* str_val)
{
  uint4_t h = 0;
  for ( ; *str_val; ++str_val) h = 5*h + *str_val;
  return  h % ht_size ;
}

#ifdef __GNUG__
template class dynarray<void*>;
#endif

