append_pfile_i::append_pfile_i(const stid_t& stid_) 
 : scan_pfile_i(stid_, t_cc_append)
{
    _init_constructor();
    INIT_SCAN_PROLOGUE_RC(append_pfile_i::append_pfile_i, 0);
    W_IGNORE(_init(true));
    if(_error_occurred) return;
	_error_occurred = lm->lock(stid, EX, t_long, WAIT_SPECIFIED_BY_XCT);
    if(_error_occurred) return;
    sdesc_t *sd;
    _error_occurred = dir->access(stid, sd, IX); 
    // makes a copy - only because the create_rec
    // functions that we want to call require that we
    // have one to reference.
    _cached_sdesc = *sd;
    w_assert3( !_page().is_fixed() );
}

append_pfile_i::append_pfile_i(const lvid_t& lvid, const serial_t& lfid)
 : scan_pfile_i(lvid, lfid, t_cc_append)
{
    _init_constructor();
    INIT_SCAN_PROLOGUE_RC(append_pfile_i::append_pfile_i, 0);
    w_rc_t rc = _init_logical(lvid, lfid);
    if (rc) return;
    W_IGNORE(_init(true));
    if(_error_occurred) return;
	_error_occurred = lm->lock(stid, EX, t_long, WAIT_SPECIFIED_BY_XCT);
    if(_error_occurred) return;

    sdesc_t *sd;
    _error_occurred = dir->access(stid, sd, IX); 
    // makes a copy - only because the create_rec
    // functions that we want to call require that we
    // have one to reference.
    _cached_sdesc = *sd;
    w_assert3( !_page().is_fixed() );
}


void
append_pfile_i::_init_constructor()
{
    // FUNC(append_pfile_i::_init_constructor);
    //  just make sure _page_alias is big enough
    w_assert3(sizeof(_page_alias) >= sizeof(pfile_p));

    new (_page_alias) pfile_p();
    w_assert3( !_page().is_fixed() );

}

append_pfile_i::~append_pfile_i() 
{ 
    FUNC(append_pfile_i::~append_pfile_i);
    if(_page().is_fixed()) {
	_page().unfix();
    }
    DBG( <<  (_page().is_fixed()? "IS FIXED-- ERROR" : "OK") );
    _page().destructor();

    // Must invalidate so that ref counts are adjusted
    DBGTHRD(<<"invalidating cached descr");
    _cached_sdesc.invalidate_sdesc();
    finish(); 
}

rc_t
append_pfile_i::next(pin_prec_i*&, smsize_t, bool& )
{
    w_assert1(0);
    return RCOK;
}

/* I don't have a create_rec with lrids yet
rc_t			
append_pfile_i::create_rec(
	const vec_t& 		    hdr,
	smsize_t 	            len_hint, 
	const vec_t& 	 	    data,
	lrid_t& 		    lrid
	)
{
    SCAN_METHOD_PROLOGUE1;
    SM_PROLOGUE_RC(append_pfile_i::create_rec, in_xct, 0);


#ifdef W_DEBUG
    if(_page().is_fixed()) {
	DBG(<<"IS FIXED! ");
    }
#endif
    serial_t serial;
    if(_error_occurred) { 
	return _error_occurred;
    }
    if( ! is_logical()) {
        return RC(eBADSCAN);
    }

    vid_t  vid;  // physical volume ID (returned by generate_new_serial)

    W_DO(lid->generate_new_serials(_lvid, vid, 1, serial, lid_m::local_ref));

    W_DO( fi->create_rec_at_end(_page(), 
	len_hint, hdr, data, serial, _cached_sdesc,  curr_rid) );


    W_DO(lid->associate(_lvid, serial, curr_rid));
    lrid.serial = serial;
    lrid.lvid = _lvid;

    return RCOK;
}
*/

rc_t			
append_pfile_i::create_rec(const precord_t& prec_info, rid_t& rid)
{
    SCAN_METHOD_PROLOGUE(append_pfile_i::create_rec);

#ifdef W_DEBUG
    if(_page().is_fixed()) {
	DBG(<<"IS FIXED! ");
    }
#endif /* W_DEBUG */
    serial_t   serial;

    if(_error_occurred) { 
	return _error_occurred;
    }
    if(is_logical()) {
        return RC(eBADSCAN);
    }

    W_DO( fi->create_rec_at_end(_page(), 
	len_hint, hdr, data, serial, _cached_sdesc,  curr_rid) );

    rid = curr_rid;
    return RCOK;
}

