/*<std-header orig-src=SHORE incl-file-exclusion=LOCK_S_H>

 $Id: lock_s.h,v 3.0 2000/09/18 21:05:12 natassa Exp $

SHORE -- Scalable Heterogeneous Object REpository

Copyright (c) 1994-98 Computer Sciences Department, University of
                      Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE AUTHORS AND THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY
OF WISCONSIN - MADISON ALLOW FREE USE OF THIS SOFTWARE IN ITS
"AS IS" CONDITION, AND THEY DISCLAIM ANY LIABILITY OF ANY KIND
FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.

This software was developed with support by the Advanced Research
Project Agency, ARPA order number 018 (formerly 8230), monitored by
the U.S. Army Research Laboratory under contract DAAB07-91-C-Q518.

*/

#ifndef LOCK_S_H
#define LOCK_S_H


/*  -- do not edit anything above this line --   </std-header>*/

#ifdef __GNUG__
#pragma interface
#endif

class lock_base_t : public smlevel_1 {
public:
    // Their order is significant.

    enum status_t {
	t_granted = 1,
	t_converting = 2,
	t_waiting = 4,
	t_aborted = 8,
	t_aborted_convert = 16
    };

    typedef lock_mode_t mode_t;

    typedef lock_duration_t duration_t;

    enum {
	MIN_MODE = NL, MAX_MODE = EX,
	NUM_MODES = MAX_MODE - MIN_MODE + 1,
    };

    static const char* const 	mode_str[NUM_MODES];
    static const char* const 	duration_str[t_num_durations];
    static const bool 		compat[NUM_MODES][NUM_MODES];
    static const mode_t 	supr[NUM_MODES][NUM_MODES];
};

#ifndef LOCK_S
/*
typedef lock_base_t::duration_t lock_duration_t;
// typedef lock_base_t::mode_t lock_mode_t; lock_mode_t defined in basics.h
typedef lock_base_t::status_t status_t;

#define LOCK_NL 	lock_base_t::NL
#define LOCK_IS 	lock_base_t::IS
#define LOCK_IX 	lock_base_t::IX
#define LOCK_SH 	lock_base_t::SH
#define LOCK_SIX	lock_base_t::SIX
#define LOCK_UD 	lock_base_t::UD
#define LOCK_EX 	lock_base_t::EX

#define LOCK_INSTANT 	lock_base_t::t_instant
#define LOCK_SHORT 	lock_base_t::t_short
#define LOCK_MEDIUM	lock_base_t::t_medium
#define LOCK_LONG 	lock_base_t::t_long
#define LOCK_VERY_LONG	lock_base_t::t_very_long
*/
#endif

class lockid_t {
public:
    NORET			W_FASTNEW_CLASS_DECL(lockid_t);
    //
    // The lock graph consists of 6 node: volumes, stores, pages, key values,
    // records, and extents. The first 5 of these form a tree of 4 levels.
    // The node for extents is not connected to the rest. The node_space_t
    // enumerator maps node types to integers. These numbers are used for
    // indexing into arrays containing node type specific info per entry (e.g
    // the lock caches for volumes, stores, and pages).
    //
    enum { NUMNODES = 6 };
    // The per-xct cache only caches volume, store, and page locks.
    enum { NUMLEVELS = 4 };
    enum name_space_t { // you cannot change these values with impunity
	t_bad		= 10,
	t_vol		= 0,
	t_store		= 1,	// parent is 1/2 = 0 t_vol
	t_page		= 2,	// parent is 2/2 = 1 t_store
	t_kvl		= 3,	// parent is 3/2 = 1 t_store
	t_record	= 4,	// parent is 4/2 = 2 t_page
	t_extent	= 5
    };

private:
    union {
	uint4_t w[4]; 
		      // w[0] == s[0,1] see below
		      // w[1] == store or extent
		      // w[2] == page
		      // w[3] contains slot (in s[6])
	uint2_t s[8]; 
		      // s[0] high byte == lspace (lock type)
		      // s[0] low byte == boolean used for extent-not-freeable
		      // s[1] vol
		      // s[2,3]==w[1] 
		      // s[4,5]== w[2] see above
		      // s[6] == slot 
		      // s[7] not used
    };

public:
    bool operator==(const lockid_t& p) const;
    bool operator!=(const lockid_t& p) const;
    friend ostream& operator<<(ostream& o, const lockid_t& i);
public:
    uint4_t 			hash() const;
    void 			zero();

private:

    //
    // lspace -- contains enum for type of lockid_t
    //
public:
    name_space_t	 	lspace() const;
private:
    void	 		set_lspace(lockid_t::name_space_t value);
    uint1_t		 	lspace_bits() const;

    //
    // vid - volume
    //
public:
    vid_t	 		vid() const;
private:
    void	 		set_vid(const vid_t & v);
    uint2_t		 	vid_bits() const;

    //
    // store - stid
    //
public:
    const snum_t&	 	store() const;
private:
    void	 		set_snum(const snum_t & s);
    void	 		set_store(const stid_t & s);
    uint4_t		 	store_bits() const;

    //
    // extent -- used only when lspace() == t_extent
    //
public:
    const extnum_t&	 	extent() const;
private:
    void	 		set_extent(const extnum_t & e);
    uint4_t	 		extent_bits() const;

    //
    // page id
    //
public:
    const shpid_t&	 	page() const;
private:
    void	 		set_page(const shpid_t & p);
    uint4_t	 		page_bits() const ;
    //
    // slot id
    //
public:
    const slotid_t&	 	slot() const;
private:
    void	 		set_slot(const slotid_t & e);
    uint2_t	 		slot_bits() const;
    uint4_t	 		slot_kvl_bits() const;

public:

    void			set_ext_has_page_alloc(bool value);
    bool			ext_has_page_alloc() const ;

    NORET			lockid_t() ;    
    NORET			lockid_t(const vid_t& vid);
    NORET			lockid_t(const extid_t& extid);    
    NORET			lockid_t(const stid_t& stid);
    NORET			lockid_t(const lpid_t& lpid);
    NORET			lockid_t(const stpgid_t& stpgid);
    NORET			lockid_t(const rid_t& rid);
    NORET			lockid_t(const kvl_t& kvl);
    NORET			lockid_t(const lockid_t& i);	

    void			extract_extent(extid_t &e) const;
    void			extract_stid(stid_t &s) const;
    void			extract_lpid(lpid_t &p) const;
    void			extract_rid(rid_t &r) const;
    void			extract_kvl(kvl_t &k) const;

    void			truncate(name_space_t space);

    lockid_t& 			operator=(const lockid_t& i);

};



#ifndef W_DEBUG
#include "lock_s_inline.h"
#endif

inline uint4_t hash(const lockid_t& id)
{
    return id.hash();
}

struct locker_mode_t {
    tid_t	tid;
    lock_mode_t	mode;
};

struct lock_info_t {
    lockid_t		name;
    lock_mode_t		group_mode;
    bool		waiting;

    lock_base_t::status_t	status;
    lock_mode_t		xct_mode;
    lock_mode_t		convert_mode;
    lock_duration_t	duration;
    int			count;
};

/*<std-footer incl-file-exclusion=LOCK_S_H>  -- do not edit anything below this line -- */

#endif          /*</std-footer>*/
