/*<std-header orig-src='shore'>

 $Id: sthread_stats.cpp,v 3.0 2000/09/18 21:05:33 natassa Exp $

SHORE -- Scalable Heterogeneous Object REpository

Copyright (c) 1994-99 Computer Sciences Department, University of
                      Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE AUTHORS AND THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY
OF WISCONSIN - MADISON ALLOW FREE USE OF THIS SOFTWARE IN ITS
"AS IS" CONDITION, AND THEY DISCLAIM ANY LIABILITY OF ANY KIND
FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.

This software was developed with support by the Advanced Research
Project Agency, ARPA order number 018 (formerly 8230), monitored by
the U.S. Army Research Laboratory under contract DAAB07-91-C-Q518.
Further funding for this work was provided by DARPA through
Rome Research Laboratory Contract No. F30602-97-2-0247.

*/

#include "w_defines.h"

/*  -- do not edit anything above this line --   </std-header>*/


#include <w_statistics.h>

#include "sthread_stats.h"
#include "sthread_stats_inc_gen.cpp"
/*
 * #include "sthread_stats_stat_gen.cpp"
 * #include "sthread_stats_out_gen.cpp"
 not used
 */

const char *sthread_stats::stat_names[] = {
#include "sthread_stats_msg_gen.h"
};

/* This replaces that generated by stats.pl */
ostream &operator <<(ostream &o, const sthread_stats &s)
{
	o << "STHREAD STATS:" << endl;
	o << "\t" << "CTX switches: " << s.ctxsw << endl;
	o << "\t" << "Idle thread:: " 
		<< s.idle_yield_return  << " yields, "
		// << s.selects  << " waits==selects() "
		<< endl;
	o << "\t" << "Spins: " << s.spins << endl;

	o << "Latches, mutexes, semaphores: "  << endl
 	  << "\tlatch waits:        " << s.latch_wait << endl
 	  << "\tlatch wait time:    " << s.latch_time << endl
 	  << "\tmutex waits:        " << s.mutex_wait << endl
 	  << "\tcondition waits:    " << s.scond_wait << endl
 	  << "\tevent sem waits:    " << s.sevsem_wait << endl
	  ;

	o << "Events:" << endl 
 	  << "\tevent waits:        " << s.selects << endl
	  << "\ttimed out:          " << s.idle << endl
	  << "\tinterrupted:        " << s.eintrs << endl
	  << "\tfound something:    " << s.selfound << endl
	  << "\tidle time in select:" << s.idle_time << endl
	  ;
	if (s.selects) {
		o << "\t#waited:           ";
		if (s.selw0)  o << " zero: " << s.selw0;
		if (s.selw1)  o << " one: " << s.selw1;
		if (s.selw2)  o << " two: " << s.selw2;
		if (s.selw3)  o << " three: " << s.selw3;
		if (s.selw4)  o << " four: " << s.selw4;
		if (s.selwX)  o << " >4: " << s.selwX;
		o << endl;
	}
	if (s.selfound) {
		o << "\t#found:            ";
		if (s.self1)  o << " one: " << s.self1;
		if (s.self2)  o << " two: " << s.self2;
		if (s.self3)  o << " three: " << s.self3;
		if (s.self4)  o << " four: " << s.self4;
		if (s.selfX)  o << " >4: " << s.selfX;
		o << endl;
	}
	o << endl;

	o << "I/O:" << endl 
	    << "\tio_time: " << s.io_time
	    << "  count: " << s.io
	    << "  average: " << (s.io ? s.io_time / s.io : 0)
	    << endl;

	o <<   "\tread: " << s.read
	    << "  write: " << s.write
	    << "  sync: " << s.sync
	    << "  trunc: " << s.truncate << endl;

	if (s.writev) {
		o << "\twritev: " << s.writev;
		if (s.writev_coalesce)
			o << " (coalesced " << s.writev_coalesce << ")";
		o << endl;
	}

	o << "\tconcurrent: " << s.ccio;
	if (s.ccio2) o << "  two: " << s.ccio2;
	if (s.ccio3) o << "  three: " << s.ccio3;
	if (s.ccio4) o << "  four: " << s.ccio4;
	if (s.ccio)
		o << "  >four: " << (s.ccio - s.ccio2 - s.ccio3 - s.ccio4);
	o  << endl;

	o << "\tlocal: " << s.local_io << endl;
	o << "\tdiskrw: " << s.diskrw_io  
	    << "   iowaits: " << s.iowaits
	    << endl;
        o <<   "\tfastpath I/O: " << s.fastpath
            << "   slow writes: " << s.nofastwrite 
	    << endl;

	o << "\tI/Os waiting the given #waits:\n\t";
	    if(s.zero>0) o << " zero: " << s.zero;
	    if(s.one>0)  o << " one: " << s.one ;
	    if(s.two>0)  o << " two: " << s.two ;
	    if(s.three>0)o << " three: " << s.three;
	    if(s.more>0)o << " >3: " << s.more;
	    if(s.wrapped>0)o << " wrapped: " << s.wrapped;
	    o << endl;

	return o;
}
